"""Tool tree helping module."""


# Due to typer usage:

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import slurmbench.topics.visitor as topics_visitor


def tool_tree(topics_type: type[topics_visitor.Topics]) -> str:
    """Tool tree command."""
    md_string = (
        "# Tool tree\n"
        "\n"
        "Topics are at the first level, and tools are at the second level.\n"
        "The commands are under parenthesis.\n"
        "\n"
    )

    for topic in topics_type:
        topic_description = topic.to_description()
        tools = topic.tools()
        md_string += f"* `{topic_description.name()}` (`{topic_description.cmd()}`)\n"
        for tool in tools:
            tool_description = tool.to_description()
            md_string += (
                f"  * `{tool_description.name()}` (`{tool_description.cmd()}`)\n"
            )

    return md_string



"""Common subprocess module."""

from __future__ import annotations

import logging
import shutil
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any

from .yaml_interface import YAMLInterface

if TYPE_CHECKING:
    from collections.abc import Sequence

_LOGGER = logging.getLogger(__name__)


def command_path(command_str: str | Path) -> Path:
    """Get command path.

    Raises
    ------
    CommandNotFoundError
        If command not found.

    """
    cmd_path = shutil.which(command_str)
    if cmd_path is None:
        _LOGGER.critical("Command not found: %s", command_str)
        raise CommandNotFoundError(command_str)
    return Path(cmd_path)


def run_cmd(cli_line: Sequence[object], cmd_str: str) -> subprocess.CompletedProcess:
    """Run external command."""
    try:
        result = subprocess.run(  # noqa: S603
            [str(x) for x in cli_line],
            check=True,
        )
    except subprocess.CalledProcessError as exc:
        _cmd_err = CommandFailedError(cmd_str, exc)
        _LOGGER.critical(str(_cmd_err))
        raise _cmd_err from exc
    return result


class CommandNotFoundError(Exception):
    """Command not found error."""

    def __init__(self, command: str | Path) -> None:
        """Initialize."""
        super().__init__()
        self.__command = command

    def __str__(self) -> str:
        """Return the error message."""
        return f"Command not found: {self.__command}"


class CommandFailedError(Exception):
    """Command failed error."""

    def __init__(
        self,
        cmd_str: str,
        called_proc_exc: subprocess.CalledProcessError,
    ) -> None:
        """Initialize."""
        super().__init__()
        self.__cmd_str = cmd_str
        self.__called_proc_exc = called_proc_exc

    def cmd_str(self) -> str:
        """Return the command string."""
        return self.__cmd_str

    def called_proc_exc(self) -> subprocess.CalledProcessError:
        """Return the command."""
        return self.__called_proc_exc

    def __str__(self) -> str:
        """Return the error message."""
        return f"{self.__cmd_str} command failed: {self.__called_proc_exc.stderr}"


class RessourcesConfig(YAMLInterface):
    """Ressources config."""

    # REFACTOR RessourcesConfig cls probably not needed

    DEFAULT_MAX_CORES = 8
    DEFAULT_MAX_MEMORY = 8

    KEY_MAX_CORES = "max_number_of_cores"
    KEY_MAX_MEMORY = "max_memory"

    @classmethod
    def from_dict(cls, config_dict: dict[str, Any]) -> RessourcesConfig:
        """Convert dict to object."""
        return cls(
            config_dict.get(cls.KEY_MAX_CORES, cls.DEFAULT_MAX_CORES),
        )

    def __init__(
        self,
        max_cores: int = DEFAULT_MAX_CORES,
        max_memory: int = DEFAULT_MAX_MEMORY,
    ) -> None:
        """Initialize object.

        Parameters
        ----------
        max_cores : int, optional
            Max number of cores, by default DEFAULT_MAX_CORES
        max_memory : int, optional
            Max memory usage (in GB), by default DEFAULT_MAX_MEMORY
        """
        self.__max_cores = max_cores
        self.__max_memory = max_memory

    def max_cores(self) -> int:
        """Get max number of cores option."""
        return self.__max_cores

    def max_memory(self) -> int:
        """Get max memory option (in GB)."""
        return self.__max_memory

    def to_dict(self) -> dict[str, Any]:
        """Convert to dict."""
        return {
            self.KEY_MAX_CORES: self.__max_cores,
            self.KEY_MAX_MEMORY: self.__max_memory,
        }



"""Environment auto documentation."""

import logging
from pathlib import Path
from typing import Annotated

import typer
from rich.markdown import Markdown as Md

import slurmbench.experiment.file_system as exp_fs
import slurmbench.tool.description as tool_desc
import slurmbench.topic.description as topic_desc
from slurmbench import root_logging
from slurmbench.tool.environments import MissingMagicCommentError, ScriptWrapper

_LOGGER = logging.getLogger(__name__)


def main(
    outdir: Annotated[Path, typer.Argument(help="Environment documentation directory")],
) -> None:
    """Give an example of using the environment wrapper."""
    toytopic_desc = topic_desc.Description("$TOPIC", "topic-cmd")
    toytool_desc = tool_desc.Description("$TOOL", "tool-cmd", toytopic_desc)
    data_exp_fs_manager = exp_fs.DataManager(
        Path("DATA_DIR"),
        toytool_desc,
        "$exp_name",
    )

    md_string = "# Example of using the environment wrapper\n\n"

    outdir.mkdir(exist_ok=True)

    md_string += "\n".join(
        [
            "",
            "## Environment config",
            "",
            "Each tool is associated with a environment context script."
            f" For a topic `{toytopic_desc.name()}`"
            f" and a tool `{toytool_desc.name()}`,"
            " the environment context script file is:"
            f" `{data_exp_fs_manager.tool_env_script_sh()}`.",
            "",
            f"The example file is in `{outdir}`",
            "",
        ],
    )

    shell_lines = "\n".join(
        [
            "# text before begin",
            "",
            f"{ScriptWrapper.BEGIN_ENV_MAGIC_COMMENT} ====",
            "",
            "source ./virtualenv_tool/bin/activate",
            "load cluster_binary/0.1.2",
            "",
            f"{ScriptWrapper.MID_ENV_MAGIC_COMMENT} ----",
            "",
            "deactivate",
            "unload cluster_binary/0.1.2",
            "",
            f"{ScriptWrapper.END_ENV_MAGIC_COMMENT} ====",
            "",
            "# text after end",
            "",
        ],
    )
    tmp_script = outdir / data_exp_fs_manager.TOOL_ENV_WRAPPER_SCRIPT_NAME
    tmp_script.write_text(shell_lines)

    md_string += "\n".join(
        [
            "",
            "## Script to describe the environment context",
            "",
            "It is constitutes in two parts:",
            "",
            "* The init part",
            "* The close part",
            "",
            "Each part is separated by magic comments.",
            "The magic comments are:",
            "",
            f"* `{ScriptWrapper.BEGIN_ENV_MAGIC_COMMENT}`",
            f"* `{ScriptWrapper.MID_ENV_MAGIC_COMMENT}`",
            f"* `{ScriptWrapper.END_ENV_MAGIC_COMMENT}`",
            "",
            f"`{tmp_script}` is an example of such script:",
            "",
            "```sh",
            shell_lines,
            "```",
            "",
        ],
    )

    bash_env = ScriptWrapper.new(tmp_script)
    if isinstance(bash_env, MissingMagicCommentError):
        _LOGGER.critical(bash_env)
        raise typer.Exit(1)

    test_wrapper_md = "\n".join(
        [
            "",
            "## Tests on the environment wrapper",
            "",
            "### Iterate over the script lines that init the environment",
            "",
            "```sh",
            "".join(bash_env.init_env_lines()) + "```",
            "",
            "### Iterate over the script lines that close the environment",
            "",
            "```sh",
            "".join(bash_env.close_env_lines()) + "```",
            "",
        ],
    )
    md_string += test_wrapper_md

    root_logging.CONSOLE.print(Md(md_string))

    readme = outdir / "README.md"
    readme.write_text(md_string)

    root_logging.CONSOLE.print(Md(f"\nThis standard output is captured in `{readme}`"))



"""Slurm file manager."""

from __future__ import annotations

from enum import StrEnum
from pathlib import Path

import slurmbench.experiment.bash.items as exp_bash_items

from . import status as slurm_status


# REFACTOR not so clear, follow work dir structure for py module(?)
class LogsManager:
    """Slurm logs manager."""

    def __init__(self, log_directory: Path) -> None:
        """Initialize."""
        self.__log_directory = log_directory
        self.__job_id_file_manager = JobIDFileManager(self.__log_directory)

    def log_dir(self) -> Path:
        """Get log directory."""
        return self.__log_directory

    def job_id_file_manager(self) -> JobIDFileManager:
        """Get array job id file manager."""
        return self.__job_id_file_manager

    def stdout(self, job_id: str) -> Path:
        """Get sbatch out file."""
        return self.log_dir() / SlurmLogFilesBuilder.io_stream(
            job_id,
            SlurmLogFilesBuilder.IOStreams.STDOUT,
        )

    def stderr(self, job_id: str) -> Path:
        """Get sbatch err file."""
        return self.log_dir() / SlurmLogFilesBuilder.io_stream(
            job_id,
            SlurmLogFilesBuilder.IOStreams.STDERR,
        )

    def script_step_status_file(
        self,
        job_id: str,
        step: exp_bash_items.Steps,
        status: slurm_status.ScriptSteps,
    ) -> Path:
        """Get bash step status file."""
        return self.log_dir() / ScriptStepStatusFilesBuilder.script_step(
            job_id,
            step,
            status,
        )

    # REFACTOR rename CommandStep by ScriptStep


class JobIDFileManager:
    """Array job ID file manager."""

    # File containing the job ID (equals to %A or SLURM_ARRAY_JOB_ID)
    ARRAY_JOB_ID_FILENAME = Path("array_job.id")

    def __init__(self, log_directory: Path) -> None:
        """Initialize."""
        self.__log_directory = log_directory

    def path(self) -> Path:
        """Get array job ID file path."""
        return self.__log_directory / self.ARRAY_JOB_ID_FILENAME

    def get_job_id(self) -> str:
        """Get array job ID."""
        return self.path().read_text().strip()


class SlurmLogFilesBuilder:
    """Slurm logs files builder."""

    class IOStreams(StrEnum):
        """IO streams."""

        STDOUT = "stdout"
        STDERR = "stderr"

    LOG_EXT = "log"

    @classmethod
    def io_stream(cls, job_id: str, stream: IOStreams) -> Path:
        """Get IO stream filename."""
        return Path(f"{job_id}_{stream}.{cls.LOG_EXT}")


class ScriptStepStatusFilesBuilder:
    """Script step status files builder."""

    # REFACTOR use common part with scripts
    class CommandStepsPrefix(StrEnum):
        """Command steps."""

        INIT_ENV = "init_env"
        COMMAND = "command"
        CLOSE_ENV = "close_env"

        @classmethod
        def from_step(
            cls,
            step: exp_bash_items.Steps,
        ) -> ScriptStepStatusFilesBuilder.CommandStepsPrefix:
            """Convert bash step to command step prefix."""
            match step:
                case exp_bash_items.Steps.INIT_ENV:
                    return cls.INIT_ENV
                case exp_bash_items.Steps.COMMAND:
                    return cls.COMMAND
                case exp_bash_items.Steps.CLOSE_ENV:
                    return cls.CLOSE_ENV

    class CommandStepsExt(StrEnum):
        """Command steps."""

        OK = "ok"
        ERROR = "error"

        @classmethod
        def from_status(
            cls,
            status: slurm_status.ScriptSteps,
        ) -> ScriptStepStatusFilesBuilder.CommandStepsExt:
            """Convert command step status to command step ext."""
            match status:
                case slurm_status.ScriptSteps.OK:
                    return cls.OK
                case slurm_status.ScriptSteps.ERROR:
                    return cls.ERROR
                case slurm_status.ScriptSteps.NULL:
                    raise ValueError

    @classmethod
    def script_step(
        cls,
        job_id: str,
        step: exp_bash_items.Steps,
        step_status: slurm_status.ScriptSteps,
    ) -> Path:
        """Get command step status filename."""
        return Path(
            f"{job_id}_{cls.CommandStepsPrefix.from_step(step)}.{cls.CommandStepsExt.from_status(step_status)}",
        )



"""Sample items."""


class RowNumbered:
    """Row numbered sample.

    The row index corresponds to the line number, where the indices is 0-based.
    So the first item row is at line number 1.
    """

    def __init__(self, row_idx: int, sample_uid: str) -> None:
        self.__row_idx = row_idx
        self.__sample_uid = sample_uid

    def row_number(self) -> int:
        """Get row number."""
        return self.__row_idx

    def uid(self) -> str:
        """Get item."""
        return self.__sample_uid

    def to_base_one(self) -> int:
        """Get line number with a base 1.

        The header is at line number 0 (base 0).
        So the first item row (base 0, row number is 1)
        is associated to line number 2 (base 1).
        """
        return self.__row_idx + 1



"""Slurm module."""

from __future__ import annotations

import logging
import subprocess
from itertools import chain
from typing import TYPE_CHECKING

import slurmbench.bash.items as bash_items
import slurmbench.experiment.file_system as exp_fs
import slurmbench.experiment.slurm.file_system as exp_slurm_fs
from slurmbench import subprocess_lib
from slurmbench.slurm import sacct

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from pathlib import Path

_LOGGER = logging.getLogger(__name__)

SBATCH_CMD = "sbatch"


def array_task_job_id(array_job_id: str, task_job_id: str) -> str:
    """Get array task job id from each job id part."""
    return f"{array_job_id}_{task_job_id}"


SLURM_ARRAY_JOB_ID_VAR = bash_items.Variable("SLURM_ARRAY_JOB_ID")
SLURM_ARRAY_TASK_ID_VAR = bash_items.Variable("SLURM_ARRAY_TASK_ID")
SLURM_JOB_ID_FROM_VARS = array_task_job_id(
    SLURM_ARRAY_JOB_ID_VAR.eval(),
    SLURM_ARRAY_TASK_ID_VAR.eval(),
)


class SbatchCommentLinesBuilder:
    """Sbatch comment lines builder."""

    COMMENT = "#SBATCH"

    ARRAY_JOB_ID = "%A"
    TASK_JOB_ID = "%a"
    JOB_ID = array_task_job_id(ARRAY_JOB_ID, TASK_JOB_ID)

    @classmethod
    def lines(
        cls,
        slurm_opts: str,
        samples_to_run_indices: Iterable[int],
        work_exp_fs_manager: exp_fs.WorkManager,
    ) -> Iterator[str]:
        """Iterate over the sbatch comment lines."""
        return (
            f"{cls.COMMENT} {line}"
            for line in chain(
                cls._job_name_lines(work_exp_fs_manager),
                cls._job_array_lines(samples_to_run_indices),
                cls._sbatch_option_log_lines(
                    work_exp_fs_manager.slurm_log_fs_manager(),
                ),
                iter((slurm_opts,)),
            )
        )

    @classmethod
    def _job_name_lines(cls, work_exp_fs_manager: exp_fs.WorkManager) -> Iterator[str]:
        """Iterate over the job name lines."""
        job_name = "_".join(
            [
                work_exp_fs_manager.tool_description().topic().name(),
                work_exp_fs_manager.tool_description().name(),
                work_exp_fs_manager.experiment_name(),
            ],
        )
        yield f"--job-name={job_name}"

    @classmethod
    def _job_array_lines(cls, samples_to_run_indices: Iterable[int]) -> Iterator[str]:
        """Iterate over the job array comment lines."""
        array_job_str = "--array=" + ",".join(
            str(sample_index) for sample_index in samples_to_run_indices
        )
        yield array_job_str

    @classmethod
    def _sbatch_option_log_lines(
        cls,
        logs_manager: exp_slurm_fs.LogsManager,
    ) -> Iterator[str]:
        """Iterate over the sbatch log option lines."""
        yield f"--output={logs_manager.stdout(cls.JOB_ID)}"
        yield f"--error={logs_manager.stderr(cls.JOB_ID)}"


SACCT_CMD = "sacct"
BASH_CMD = "bash"


def write_slurm_stats(job_id: str, psv_path: Path) -> None:
    """Write sbatch stats."""
    sacct_cmd = f"{SACCT_CMD} --long --jobs {job_id} --parsable2 > {psv_path}"

    cmd_path = subprocess_lib.command_path(BASH_CMD)
    result = subprocess.run(  # noqa: S603
        [str(x) for x in [cmd_path, "-c", f"'{sacct_cmd}'"]],
        capture_output=True,
        text=True,
        check=False,
    )
    _LOGGER.debug("%s stdout: %s", SACCT_CMD, result.stdout)
    _LOGGER.debug("%s stderr: %s", SACCT_CMD, result.stderr)


def get_state(job_id: str) -> sacct.State | None:
    """Get job state if exists for the job id."""
    return get_states([job_id]).get(job_id, None)


def get_states(job_ids: Iterable[str]) -> dict[str, sacct.State]:
    """Get jobs states.

    Warning
    -------
    The job id can be absent from the sacct output.
    The user must verify if the job id is in the dictionary keys.
    """
    cmd_path = subprocess_lib.command_path(SACCT_CMD)
    result = subprocess.run(  # noqa: S602
        f"{cmd_path} --jobs " + ",".join(job_ids) + " --format=JobID,State -P -X",
        shell=True,
        capture_output=True,
        text=True,
        check=False,
    )
    job_states: dict[str, sacct.State] = {}
    for stdout_line in result.stdout.splitlines()[1:]:
        columns = stdout_line.split("|")
        job_states[columns[0]] = sacct.State(columns[1])
    return job_states



"""Tools script logics."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, final

import slurmbench.bash.items as bash_items
import slurmbench.experiment.file_system as exp_fs
import slurmbench.samples.bash as smp_sh

from . import results

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator
    from pathlib import Path


class Argument[R: results.Result]:
    """Argument bash lines builder."""

    def __init__(
        self,
        input_result: R,
        work_exp_fs_manager: exp_fs.WorkManager,
    ) -> None:
        """Initialize."""
        self._input_result = input_result
        self._input_data_smp_sh_fs_manager = smp_sh.sample_shell_fs_manager(
            input_result.exp_fs_manager(),
        )
        self._work_exp_fs_manager = work_exp_fs_manager
        self._work_smp_sh_fs_manager = smp_sh.sample_shell_fs_manager(
            self._work_exp_fs_manager,
        )

    def input_result(self) -> R:
        """Get input result."""
        return self._input_result

    def input_data_smp_sh_fs_manager(self) -> smp_sh.smp_fs.Manager:
        """Get input data sample shell file system manager."""
        return self._input_data_smp_sh_fs_manager

    def work_exp_fs_manager(self) -> exp_fs.WorkManager:
        """Get working experiment file system manager."""
        return self._work_exp_fs_manager

    def work_smp_sh_fs_manager(self) -> smp_sh.smp_fs.Manager:
        """Get working sample shell file system manager."""
        return self._work_smp_sh_fs_manager

    def init_lines(self) -> Iterator[str]:
        """Get shell input init lines."""
        yield from ()

    def close_lines(self) -> Iterator[str]:
        """Get shell input close lines."""
        yield from ()


class Options:
    """Bash lines builder for user tool options."""

    USER_TOOL_OPTIONS_VAR = bash_items.Variable("USER_TOOL_OPTIONS")

    def __init__(self, tool_options: Iterable[str]) -> None:
        """Initialize."""
        self.__tool_options = tool_options

    def tool_options(self) -> Iterable[str]:
        """Get tool options."""
        return self.__tool_options

    def set_options(self) -> Iterator[str]:
        """Set user tool options sh array variable."""
        yield self.USER_TOOL_OPTIONS_VAR.set(
            "(" + " ".join(self.__tool_options) + ")",
        )


class WithOptions(ABC):
    """Commands with options."""

    WORK_EXP_SAMPLE_DIR_VAR = bash_items.Variable("WORK_EXP_SAMPLE_DIR")

    CORE_COMMAND_SH_FILENAME = "core_command.sh"

    @classmethod
    def core_command_sh_path(cls, tool_bash_script_dir: Path) -> Path:
        """Get core command shell path."""
        return tool_bash_script_dir / cls.CORE_COMMAND_SH_FILENAME

    def __init__(
        self,
        opts_sh_lines_builder: Options,
        exp_fs_managers: exp_fs.Managers,
        tool_bash_script_dir: Path,
    ) -> None:
        """Initialize."""
        self._opts_sh_lines_builder = opts_sh_lines_builder
        self._exp_fs_managers = exp_fs_managers
        self._tool_bash_script_dir = tool_bash_script_dir

    @abstractmethod
    def init_lines(self) -> Iterator[str]:
        """Iterate over core command shell input init lines."""
        raise NotImplementedError

    @abstractmethod
    def close_lines(self) -> Iterator[str]:
        """Iterate over core command shell input close lines."""
        raise NotImplementedError

    def commands(self) -> Iterator[str]:
        """Iterate over the tool commands."""
        # DOCU say WORK_EXP_SAMPLE_DIR variable is set
        # DOCU say SAMPLES_TSV variable is set
        yield from smp_sh.SampleUIDLinesBuilder(
            self._exp_fs_managers.data().samples_tsv(),
        ).lines()
        yield ""
        yield from self.set_work_sample_exp_dir()
        yield from self.mkdir_work_sample_exp_dir()
        yield ""
        yield from self._opts_sh_lines_builder.set_options()
        yield ""
        yield from self.init_lines()
        yield ""
        yield from self.core_commands()
        yield ""
        yield from self.close_lines()

    def opts_sh_lines_builder(self) -> Options:
        """Get options bash lines builder."""
        return self._opts_sh_lines_builder

    def exp_fs_managers(self) -> exp_fs.Managers:
        """Get experiment file system managers."""
        return self._exp_fs_managers

    def set_work_sample_exp_dir(self) -> Iterator[str]:
        """Set working experiment sample directory."""
        work_exp_sample_dir = smp_sh.sample_shell_fs_manager(
            self._exp_fs_managers.work(),
        ).sample_dir()
        yield self.WORK_EXP_SAMPLE_DIR_VAR.set(
            bash_items.path_to_str(work_exp_sample_dir),
        )

    def mkdir_work_sample_exp_dir(self) -> Iterator[str]:
        """Mkdir working experiment sample directory."""
        yield f"mkdir -p {self.WORK_EXP_SAMPLE_DIR_VAR.eval()} 2>/dev/null"

    def core_commands(self) -> Iterator[str]:
        """Iterate over the tool command lines."""
        core_command_shell_path = self.core_command_sh_path(self._tool_bash_script_dir)

        with core_command_shell_path.open("r") as in_core_cmd:
            for line in in_core_cmd:
                yield line.rstrip()


class OnlyOptions(WithOptions):
    """Tool commands when the tool has no arguments."""

    def init_lines(self) -> Iterator[str]:
        """Iterate over core command shell input init lines."""
        yield from ()

    def close_lines(self) -> Iterator[str]:
        """Iterate over core command shell input close lines."""
        yield from ()


@final
class WithArguments(WithOptions):
    """Tool commands with options and arguments."""

    def __init__(
        self,
        arg_sh_lines_builders: Iterable[Argument],
        opts_sh_lines_builder: Options,
        exp_fs_managers: exp_fs.Managers,
        _tool_bash_script_dir: Path,
    ) -> None:
        """Initialize."""
        self._arg_sh_lines_builders = list(arg_sh_lines_builders)
        super().__init__(opts_sh_lines_builder, exp_fs_managers, _tool_bash_script_dir)

    def arg_sh_lines_builders(self) -> Iterator[Argument]:
        """Get argument bash lines builders."""
        yield from self._arg_sh_lines_builders

    def init_lines(self) -> Iterator[str]:
        """Iterate over core command shell input init lines."""
        for result_lines_builder in self._arg_sh_lines_builders:
            yield from result_lines_builder.init_lines()

    def close_lines(self) -> Iterator[str]:
        """Iterate over core command shell input close lines."""
        for result_lines_builder in self._arg_sh_lines_builders:
            yield from result_lines_builder.close_lines()



"""Topic results prelude."""

# Due to typer usage:
# ruff: noqa: PLC0414

from slurmbench.tool.results import Formatted as Formatted
from slurmbench.tool.results import Original as Original
from slurmbench.topic.results import ConvertFn as ConvertFn
from slurmbench.topic.results import Error as Error
from slurmbench.topic.results import FormattedVisitor as FormattedVisitor
from slurmbench.topic.results import OriginalVisitor as OriginalVisitor



"""Topic abstract application module."""

# Due to typer usage:
# ruff: noqa: TC001, TC003

from __future__ import annotations

import logging
from collections.abc import Iterable
from typing import Self

import typer

import slurmbench.topic.visitor as topic_visitor

from . import description as desc

_LOGGER = logging.getLogger(__name__)


class Topic[ToolsT: type[topic_visitor.Tools]]:
    """Topic application."""

    @classmethod
    def new(
        cls,
        topic_description: desc.Description,
        tools: ToolsT,
        tool_apps: Iterable[typer.Typer],
    ) -> Self:
        """Build topic application."""
        tool_apps = list(tool_apps)
        if len(tool_apps) != len(tools):
            _LOGGER.critical(
                "Number of tool applications (%d)"
                " differs from number of tools (%d) for topic `%s`",
                len(tool_apps),
                len(tools),
                topic_description.name(),
            )
            raise typer.Exit(1)
        app = typer.Typer(
            name=topic_description.cmd(),
            help=f"Subcommand for topic `{topic_description.name()}`",
            rich_markup_mode="rich",
        )
        for tool_app in tool_apps:
            app.add_typer(tool_app)
        return cls(tools, app)

    def __init__(self, tools: ToolsT, app: typer.Typer) -> None:
        self._tools = tools
        self._app = app

    def tools(self) -> ToolsT:
        """Get tools."""
        return self._tools

    def app(self) -> typer.Typer:
        """Get app."""
        return self._app



"""Sample experiment status logics module."""

from __future__ import annotations

from contextlib import suppress
from enum import StrEnum
from typing import TYPE_CHECKING, Self, assert_never

from slurmbench.slurm import sacct

if TYPE_CHECKING:
    import slurmbench.samples.file_system as smp_fs


class Success(StrEnum):
    """Sample experiment OK status."""

    OK = "ok"


class Error(StrEnum):
    """Sample experiment error status."""

    # The sample experiment has never been run or exit before log
    NOT_RUN = "not_run"
    # One of the input is missing
    MISSING_INPUTS = "missing_inputs"
    # An error occur during the sample experiment run or sacct state is unknown
    ERROR = "error"


type Status = Success | Error


class StatusMap[T]:
    """Status map."""

    @classmethod
    def default(cls, default_value: T) -> Self:
        """Get default status map."""
        return cls(default_value, default_value, default_value, default_value)

    def __init__(self, ok: T, missing_inputs: T, error: T, not_run: T) -> None:
        self._ok = ok
        self._missing_inputs = missing_inputs
        self._error = error
        self._not_run = not_run

    def get(self, status: Status) -> T:
        """Get status map value."""
        match status:
            case Success.OK:
                return self._ok
            case Error.MISSING_INPUTS:
                return self._missing_inputs
            case Error.ERROR:
                return self._error
            case Error.NOT_RUN:
                return self._not_run

    def set(self, status: Status, value: T) -> None:
        """Set status map value."""
        match status:
            case Success.OK:
                self._ok = value
            case Error.MISSING_INPUTS:
                self._missing_inputs = value
            case Error.ERROR:
                self._error = value
            case Error.NOT_RUN:
                self._not_run = value
            case _:
                assert_never(status)

    def __getitem__(self, status: Status) -> T:
        """Get status map value."""
        return self.get(status)

    def __setitem__(self, status: Status, value: T) -> None:
        """Set status map value."""
        self.set(status, value)


def from_str(status_str: str) -> Status:
    """Get sample experiment status from string."""
    with suppress(ValueError):
        return Error(status_str)
    return Success(status_str)


def get_status(sample_fs_manager: smp_fs.Manager) -> Status:
    """Get sample experiment status.

    Warning
    -------
    Should be use after the sample experiment is finished (so in data)
    # REFACTOR Differentiate data and work sample fs managers
    """
    if not sample_fs_manager.sample_dir().exists():
        return Error.NOT_RUN
    if sample_fs_manager.missing_inputs_tsv().exists():
        return Error.MISSING_INPUTS
    if sample_fs_manager.errors_log().exists():
        return Error.ERROR
    if sample_fs_manager.done_log().exists():
        return Success.OK
    return Error.NOT_RUN


def from_sacct_state(status: sacct.State) -> Status:
    """Get sample experiment status from sacct state."""
    match status:
        case (
            sacct.State.BOOT_FAIL
            | sacct.State.CANCELLED
            | sacct.State.DEADLINE
            | sacct.State.FAILED
            | sacct.State.NODE_FAIL
            | sacct.State.OUT_OF_MEMORY
            | sacct.State.REVOKED
            | sacct.State.TIMEOUT
        ):
            return Error.ERROR
        case sacct.State.COMPLETED:
            return Success.OK
        case (
            sacct.State.PENDING
            | sacct.State.PREEMPTED
            | sacct.State.RUNNING
            | sacct.State.REQUEUED
            | sacct.State.RESIZING
            | sacct.State.SUSPENDED
        ):
            return Error.NOT_RUN



"""Tool results prelude."""

# Due to typer usage:
# ruff: noqa: PLC0414, F401

from slurmbench.experiment.file_system import DataManager as ExpFSDataManager
from slurmbench.samples.items import RowNumbered as Sample
from slurmbench.samples.status import Error as SampleError
from slurmbench.samples.status import Status as SampleStatus
from slurmbench.samples.status import Success as SampleSuccess
from slurmbench.tool.connector import InvalidToolNameError as InvalidToolNameError
from slurmbench.tool.connector import get_arg as get_arg
from slurmbench.tool.results import Formatted as Formatted
from slurmbench.tool.results import Original as Original



"""SLURM applications."""

# Due to typer usage:

from __future__ import annotations

from typing import Annotated

import typer

from . import config


class Arguments:
    """SLURM app arguments."""

    ACCOUNT_NAME = typer.Argument(
        help="SLURM account name",
    )


def slurm_opts(
    account_name: Annotated[str | None, Arguments.ACCOUNT_NAME] = None,
) -> str:
    """Get default SLURM options string."""
    options_str = config.default_slurm_options(account_name)
    typer.echo(options_str)
    return options_str



"""YAML interface module."""

from __future__ import annotations

from abc import ABC, abstractmethod
from pathlib import Path

import yaml  # type: ignore[import-untyped]

try:
    from yaml import CDumper as Dumper
except ImportError:
    from yaml import Dumper

from typing import Self


class YAMLInterface(ABC):
    """YAML interface."""

    @classmethod
    def from_yaml(cls, yaml_filepath: Path) -> Self:
        """Get object from YAML file."""
        with yaml_filepath.open("r") as file:
            obj_dict = yaml.safe_load(file)
        return cls.from_yaml_load(obj_dict)

    @classmethod
    @abstractmethod
    def from_yaml_load(cls, pyyaml_obj: yaml.YAMLObject) -> Self:
        """Convert pyyaml object to self."""
        raise NotImplementedError

    @abstractmethod
    def to_yaml_dump(self) -> yaml.YAMLObject:
        """Convert to dict."""
        raise NotImplementedError

    def to_yaml(self, yaml_filepath: Path) -> Path:
        """Write to yaml."""
        yaml_filepath = Path(yaml_filepath)
        with yaml_filepath.open("w") as file:
            yaml.dump(self.to_yaml_dump(), file, Dumper=Dumper, sort_keys=False)
        return yaml_filepath



"""Common logging module."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import typer
from rich.console import Console
from rich.logging import RichHandler

if TYPE_CHECKING:
    from pathlib import Path

__ROOT_APP_NAME = "slurmbench"

_LOGFORMAT = "%(asctime)s %(name)s [%(levelname)s] %(message)s"
_LOGFORMAT_RICH = "%(message)s"

_LOGGER = logging.getLogger(__ROOT_APP_NAME)
_LOGGER.setLevel(logging.DEBUG)

CONSOLE = Console()

OPT_DEBUG = typer.Option(
    help="Debug mode",
)


def format_logger(
    debug: bool,  # noqa: FBT001
    log_file: Path | None,
) -> None:
    """Format logger."""
    _LOGGER.handlers.clear()
    _LOGGER.filters.clear()

    log_filter = logging.Filter(__ROOT_APP_NAME)

    _LOGGER.addFilter(log_filter)
    rich_handler = RichHandler(console=CONSOLE)

    if debug:
        rich_handler.setLevel(logging.DEBUG)
    else:
        rich_handler.setLevel(logging.INFO)
    rich_formatter = logging.Formatter(_LOGFORMAT_RICH, datefmt="%Y-%m-%dT%H:%M:%S%z")
    rich_handler.setFormatter(rich_formatter)
    _LOGGER.addHandler(rich_handler)

    if log_file is not None:
        file_handler = logging.FileHandler(log_file)
        if debug:
            file_handler.setLevel(logging.DEBUG)
        else:
            file_handler.setLevel(logging.INFO)
        file_formatter = logging.Formatter(_LOGFORMAT, datefmt="%Y-%m-%dT%H:%M:%S%z")
        file_handler.setFormatter(file_formatter)
        _LOGGER.addHandler(file_handler)


def init_logger(
    logger: logging.Logger,
    first_info_message: str,
    debug: bool,  # noqa: FBT001
    log_file: Path | None = None,
) -> None:
    """Initialize logger."""
    format_logger(debug, log_file)
    logger.info(first_info_message)



"""Sample file system manager."""

from __future__ import annotations

import shutil
from enum import StrEnum
from pathlib import Path
from typing import TYPE_CHECKING, final

from slurmbench import tab_files

from . import items as smp
from .slurm import file_system as slurm_fs

if TYPE_CHECKING:
    from collections.abc import Iterator


class Manager:
    """Sample file system manager."""

    MISSING_INPUTS_TSV_NAME = Path("missing_inputs.tsv")
    ERRORS_LOG_NAME = Path("errors.log")
    DONE_LOG_NAME = Path("done.log")

    @classmethod
    def from_exp_dir_and_sample(
        cls,
        exp_dir: Path,
        sample: smp.RowNumbered,
    ) -> Manager:
        """Get sample experiment file system manager."""
        return cls(exp_dir / sample.uid())

    def __init__(self, sample_dir: Path) -> None:
        """Inititialize."""
        self.__sample_dir = sample_dir
        self.__slurm_fs_manager = slurm_fs.Manager(self.__sample_dir)

    def sample_dir(self) -> Path:
        """Get sample directory path."""
        return self.__sample_dir

    def missing_inputs_tsv(self) -> Path:
        """Get missing_inputs file path."""
        return self.__sample_dir / self.MISSING_INPUTS_TSV_NAME

    def errors_log(self) -> Path:
        """Get errors file."""
        return self.__sample_dir / self.ERRORS_LOG_NAME

    def done_log(self) -> Path:
        """Get done file."""
        return self.__sample_dir / self.DONE_LOG_NAME

    def slurm_fs_manager(self) -> slurm_fs.Manager:
        """Get slurm file system manager."""
        return self.__slurm_fs_manager


def reset_sample_dir(manager: Manager) -> None:
    """Reset sample directory."""
    shutil.rmtree(manager.sample_dir(), ignore_errors=True)
    manager.sample_dir().mkdir(parents=True, exist_ok=True)


class TSVHeader(StrEnum):
    """TSV header."""

    UID = "uid"


@final
class TSVReader(tab_files.TSVReader[TSVHeader, smp.RowNumbered]):
    """Samples TSV reader."""

    @classmethod
    def header_type(cls) -> type[TSVHeader]:
        """Get header type."""
        return TSVHeader

    def __iter__(self) -> Iterator[smp.RowNumbered]:
        """Iterate sequence probability scores."""
        for row_idx, row in enumerate(self._csv_reader, start=1):
            yield smp.RowNumbered(row_idx, self._get_cell(row, TSVHeader.UID))


def columns_name_index(file: Path) -> dict[str, int]:
    """Get columns name index."""
    with TSVReader.open(file) as reader:
        return reader.header_map().indices()



"""Bash script creation module."""

from collections.abc import Iterable
from pathlib import Path

import slurmbench.bash.items as bash_items


def create(bash_lines: Iterable[str], script_path: Path) -> None:
    """Create a bash script.

    The bash lines should not contain a newline at the end.
    """
    with script_path.open("w") as f_out:
        f_out.write(bash_items.BASH_SHEBANG + "\n")
        f_out.write("\n")
        for line in bash_lines:
            f_out.write(line + "\n")



"""Resume experiment."""

from __future__ import annotations

from . import managers, monitors, resolve


def resume(
    exp_manager: managers.OnlyOptions | managers.WithArguments,
    array_job_id: str,
) -> None:
    """Resume experiment."""
    unresolved_samples = list(
        monitors.iter_unresolved_samples(exp_manager.fs_managers().work()),
    )
    resolve.resolve_running_samples(exp_manager, unresolved_samples, array_job_id)



"""Slurm status logics."""

from enum import StrEnum


class State(StrEnum):
    """Slurm states recognized by sacct.

    From: https://slurm.schedmd.com/sacct.html#SECTION_JOB-STATE-CODES
    """

    BOOT_FAIL = "BOOT_FAIL"
    CANCELLED = "CANCELLED"
    COMPLETED = "COMPLETED"
    DEADLINE = "DEADLINE"
    FAILED = "FAILED"
    NODE_FAIL = "NODE_FAIL"
    OUT_OF_MEMORY = "OUT_OF_MEMORY"
    PENDING = "PENDING"
    PREEMPTED = "PREEMPTED"
    RUNNING = "RUNNING"
    REQUEUED = "REQUEUED"
    RESIZING = "RESIZING"
    REVOKED = "REVOKED"
    SUSPENDED = "SUSPENDED"
    TIMEOUT = "TIMEOUT"



"""ABC for apps."""

# Due to typer usage:

from __future__ import annotations

import logging
from enum import StrEnum
from typing import TYPE_CHECKING

import typer

from .help import app as help_app
from .slurm import app as slurm_app

if TYPE_CHECKING:
    from collections.abc import Iterable

    from .topic import app as topic_app
    from .topics import visitor as topics_visitor

_LOGGER = logging.getLogger(__name__)


class CommandCategories(StrEnum):
    """Command categories."""

    TOPICS = "Topics"
    UTILITIES = "Utilities"


def new(
    name: str,
    description: str,
    topics_type: type[topics_visitor.Topics],
    topic_apps: Iterable[topic_app.Topic],
) -> typer.Typer:
    """Create a new benchmark application."""
    topic_apps = list(topic_apps)
    if len(topic_apps) != len(topics_type):
        _LOGGER.critical(
            "Number of topic applications (%d) differs from number of topics (%d)",
            len(topic_apps),
            len(topics_type),
        )
        raise typer.Exit(1)
    app = typer.Typer(
        name=name,
        help=description,
        rich_markup_mode="rich",
    )
    #
    # Utilities
    #
    app.add_typer(
        help_app.new(topics_type),
        rich_help_panel=CommandCategories.UTILITIES,
    )
    app.command(rich_help_panel=CommandCategories.UTILITIES)(slurm_app.slurm_opts)
    #
    # Topics
    #
    for sub_app in topic_apps:
        app.add_typer(sub_app.app(), rich_help_panel=CommandCategories.TOPICS)
    return app


APP = typer.Typer(
    name="slurmbench",
    help="SLURM benchmark interface",
    rich_markup_mode="rich",
)



"""Meta logic on module."""

from pathlib import Path

from .tool import description as desc
from .topic import description as topic_desc

SLURMBENCH_ROOT = Path(__file__).parent

TOPICS_MODULE_NAME = "topics"


def topic_module_path(topic_name: str) -> Path:
    """Get topic module path."""
    return SLURMBENCH_ROOT / TOPICS_MODULE_NAME / topic_name.lower()


def topic_module_path_from_description(topic: topic_desc.Description) -> Path:
    """Get topic module path."""
    return topic_module_path(topic.name())


def tool_module_path(topic_name: str, tool_name: str) -> Path:
    """Get tool module path."""
    return topic_module_path(topic_name) / tool_name.lower()


def tool_module_path_from_descriptions(
    topic: topic_desc.Description,
    tool: desc.Description,
) -> Path:
    """Get tool module path."""
    return tool_module_path(topic.name(), tool.name())


def path_to_python_path(path: Path) -> str:
    """Get python path."""
    return str(path).replace("/", ".")



"""Samples prelude."""

# ruff: noqa: PLC0414

# DOCU user can use the sample prelude

from slurmbench.samples.file_system import TSVHeader as TSVHeader
from slurmbench.samples.status import Status as Status



"""Experiment managers."""

from __future__ import annotations

from typing import TYPE_CHECKING, final

import slurmbench.tool.connector as tool_connector

if TYPE_CHECKING:
    from . import file_system as exp_fs


class WithOptions[C: tool_connector.WithOptions]:
    """Experiment manager."""

    def __init__(
        self,
        exp_fs_managers: exp_fs.Managers,
        tool_connector: C,
    ) -> None:
        self._fs_managers = exp_fs_managers
        self._tool_connector = tool_connector

    def exp_name(self) -> str:
        """Get experiment name."""
        return self._fs_managers.data().experiment_name()

    def fs_managers(self) -> exp_fs.Managers:
        """Get experiment file system managers."""
        return self._fs_managers

    def tool_connector(self) -> C:
        """Get tool connector."""
        return self._tool_connector


@final
class OnlyOptions(WithOptions[tool_connector.OnlyOptions]):
    """Experiment manager with only options."""


@final
class WithArguments[N: tool_connector.Names](
    WithOptions[tool_connector.WithArguments[N]],
):
    """Experiment manager with arguments."""



"""Tool connector module."""

from __future__ import annotations

import inspect
import logging
from abc import ABC, abstractmethod
from enum import StrEnum
from pathlib import Path
from typing import TYPE_CHECKING, Self, TypeVar, final

import slurmbench.experiment.file_system as exp_fs
import slurmbench.topic.results as topic_res
import slurmbench.topic.visitor as topic_visitor

from . import bash, results
from . import config as cfg
from . import description as desc

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator

_LOGGER = logging.getLogger(__name__)


class Names(StrEnum):
    """Tool names."""

    @abstractmethod
    def topic_tools(self) -> type[topic_visitor.Tools]:
        """Get topic tools."""
        raise NotImplementedError


class InvalidToolNameError[N: Names, T: topic_visitor.Tools]:
    """Invalid tool name error."""

    def __init__(
        self,
        arg_name: N,
        invalid_tool_name: str,
        valid_tools: Iterable[T],
    ) -> None:
        self._arg_name = arg_name
        self._invalid_tool_name = invalid_tool_name
        self._valid_tools = tuple(valid_tools)

    def arg_name(self) -> N:
        """Get argument name."""
        return self._arg_name

    def invalid_tool_name(self) -> str:
        """Get invalid tool name."""
        return self._invalid_tool_name

    def valid_tools(self) -> tuple[T, ...]:
        """Get list of valid tools."""
        return self._valid_tools


class Arg[N: Names, T: topic_visitor.Tools, R: results.Result](ABC):
    """Tool argument configuration."""

    @classmethod
    def config_type(cls) -> type[cfg.Arg]:
        """Get config type."""
        return cfg.Arg

    @classmethod
    @abstractmethod
    def name(cls) -> N:
        """Get name."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def tools_type(cls) -> type[T]:
        """Get tools type."""
        raise NotImplementedError

    @classmethod
    def valid_tools(cls) -> Iterable[T]:
        """Get valid tools."""
        return (
            tool
            for tool in cls.tools_type()
            if cls.result_visitor().tool_gives_the_result(tool)
        )

    @classmethod
    @abstractmethod
    def result_visitor(cls) -> type[topic_res.Visitor[T, R]]:
        """Get result visitor function."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def sh_lines_builder_type(cls) -> type[bash.Argument[R]]:
        """Get shell lines builder."""
        raise NotImplementedError

    @classmethod
    def from_config(cls, config: cfg.Arg) -> Self | InvalidToolNameError[N, T]:
        """Convert dict to object."""
        try:
            tool = cls.tools_type()(config.tool_name())
        except ValueError:
            return InvalidToolNameError(
                cls.name(),
                config.tool_name(),
                cls.valid_tools(),
            )
        match cls.result_visitor().result_builder_from_tool(tool):
            case topic_res.Error():
                return InvalidToolNameError(
                    cls.name(),
                    config.tool_name(),
                    cls.valid_tools(),
                )
        return cls(tool, config.exp_name())

    def __init__(self, tool: T, exp_name: str) -> None:
        """Initialize."""
        self._tool = tool
        self._exp_name = exp_name

    def tool(self) -> T:
        """Get tool."""
        return self._tool

    def exp_name(self) -> str:
        """Get experiment name."""
        return self._exp_name

    def result(self, data_exp_fs_manager: exp_fs.DataManager) -> R:
        """Convert argument to input."""
        return self.result_visitor().result_builder()(
            exp_fs.WorkManager(
                data_exp_fs_manager.root_dir(),
                self._tool.to_description(),
                self._exp_name,
            ),
        )

    def sh_lines_builder(self, exp_fs_managers: exp_fs.Managers) -> bash.Argument[R]:
        """Convert input to shell lines builder."""
        return self.sh_lines_builder_type()(
            self.result(exp_fs_managers.data()),
            exp_fs_managers.work(),
        )

    def to_config(self) -> cfg.Arg:
        """Convert to config."""
        return cfg.Arg(str(self._tool), self._exp_name)


class MissingArgumentNameError[N: Names]:
    """Missing argument name error."""

    def __init__(self, missing_arg_name: N, names_type: type[N]) -> None:
        self._missing_arg_name = missing_arg_name
        self._names_type = names_type

    def missing_arg_name(self) -> N:
        """Get missing argument name."""
        return self._missing_arg_name

    def names_type(self) -> type[N]:
        """Get names type."""
        return self._names_type


class ExtraArgumentNameError[N: Names]:
    """Extra argument name error."""

    def __init__(self, extra_arg_names: Iterable[str], names_type: type[N]) -> None:
        self._extra_arg_names = tuple(extra_arg_names)
        self._names_type = names_type

    def extra_arg_names(self) -> tuple[str, ...]:
        """Get extra argument name."""
        return self._extra_arg_names

    def names_type(self) -> type[N]:
        """Get names type."""
        return self._names_type


ArgsLoadError = InvalidToolNameError | MissingArgumentNameError | ExtraArgumentNameError

NamesTypeVar = TypeVar("NamesTypeVar", bound=Names)
ArgWithName = Arg[
    NamesTypeVar,
    topic_visitor.Tools,
    results.Result,
]


class Arguments[N: Names](ABC):
    """Tool arguments configuration."""

    @classmethod
    def config_type(cls) -> type[cfg.Arguments]:
        """Get config type."""
        return cfg.Arguments

    @classmethod
    @abstractmethod
    def arg_types(cls) -> list[type[ArgWithName[N]]]:
        """Get argument type."""
        raise NotImplementedError

    @classmethod
    def names_type(cls) -> type[N]:
        """Get names type."""
        return type(cls.arg_types()[0].name())

    @classmethod
    def from_config(cls, config: cfg.Arguments) -> Self | ArgsLoadError:
        """Convert dict to object."""
        arg_dict: dict[N, ArgWithName[N]] = {}
        for arg_type in cls.arg_types():
            try:
                arg_config = config[str(arg_type.name())]
            except KeyError:
                return MissingArgumentNameError(arg_type.name(), cls.names_type())

            match arg_or_err := arg_type.from_config(arg_config):
                case Arg():
                    arg_dict[arg_type.name()] = arg_or_err
                case InvalidToolNameError():
                    return arg_or_err

        if extra_arg := set(config.arguments().keys()) - {str(n) for n in arg_dict}:
            return ExtraArgumentNameError(extra_arg, cls.names_type())

        return cls(arg_dict)

    def __init__(self, arguments: dict[N, ArgWithName[N]]) -> None:
        self.__arguments = arguments

    def __getitem__(self, name: N) -> ArgWithName[N]:
        """Get argument."""
        return self.__arguments[name]

    def __iter__(self) -> Iterator[tuple[N, ArgWithName[N]]]:
        """Iterate arguments."""
        return iter(self.__arguments.items())

    def results(
        self,
        data_exp_fs_manager: exp_fs.DataManager,
    ) -> Iterator[tuple[N, results.Result]]:
        """Iterate over results associated with the arguments."""
        yield from (
            (name, arg.result(data_exp_fs_manager))
            for name, arg in self.__arguments.items()
        )

    def sh_lines_builders(
        self,
        exp_fs_managers: exp_fs.Managers,
    ) -> Iterator[bash.Argument]:
        """Convert to commands."""
        return (
            arg.sh_lines_builder(exp_fs_managers) for arg in self.__arguments.values()
        )

    def to_config(self) -> cfg.Arguments:
        """Convert to config."""
        return cfg.Arguments(
            {str(name): arg.to_config() for name, arg in self.__arguments.items()},
        )


@final
class StringOpts:
    """String options.

    When the options are regular short/long options.
    """

    @classmethod
    def from_config(cls, config: cfg.StringOpts) -> Self:
        """Convert dict to object."""
        return cls(config)

    def __init__(self, options: Iterable[str]) -> None:
        self.__options = list(options)

    def __bool__(self) -> bool:
        """Check if options are not empty."""
        return len(self.__options) > 0

    def __len__(self) -> int:
        """Get options length."""
        return len(self.__options)

    def __iter__(self) -> Iterator[str]:
        """Iterate options."""
        return iter(self.__options)

    def sh_lines_builder(self) -> bash.Options:
        """Get shell lines builder type."""
        return bash.Options(self)

    def to_config(self) -> cfg.StringOpts:
        """Convert to config."""
        return cfg.StringOpts(self.__options)


class WithOptions[C: cfg.WithOptions, E](ABC):
    """Tool config with options."""

    @classmethod
    @abstractmethod
    def description(cls) -> desc.Description:
        """Get tool description."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def config_type(cls) -> type[C]:
        """Get config type."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_config(cls, config: C) -> Self | E:
        """Convert dict to object."""
        raise NotImplementedError

    def __init__(self, options: StringOpts) -> None:
        """Initialize."""
        self._options = options

    def options(self) -> StringOpts:
        """Get options."""
        return self._options

    @abstractmethod
    def to_config(self) -> cfg.WithOptions:
        """Convert to config."""
        raise NotImplementedError

    @abstractmethod
    def sh_commands(
        self,
        exp_fs_managers: exp_fs.Managers,
    ) -> bash.WithOptions:
        """Get sh commands."""
        raise NotImplementedError

    def is_same(self, other: Self) -> bool:
        """Check if configs are the same."""
        return self.to_config().is_same(other.to_config())

    @classmethod
    def parent_dir_where_defined(cls) -> Path:
        """Get the parent directory of the module defining the connector.

        Usefull to retrieve the tool template bash script.
        """
        cls_mod = inspect.getmodule(cls)
        match cls_mod:
            case None:
                _LOGGER.critical("No module for %s", cls)
                raise ValueError
        if cls_mod.__file__ is None:
            _LOGGER.critical("No file for %s", cls)
            raise ValueError
        return Path(cls_mod.__file__).parent


class OnlyOptions(WithOptions[cfg.OnlyOptions, "OnlyOptions"]):
    """Tool config without arguments."""

    @classmethod
    def config_type(cls) -> type[cfg.OnlyOptions]:
        """Get config type."""
        return cfg.OnlyOptions

    @classmethod
    def from_config(cls, config: cfg.OnlyOptions) -> Self:
        """Convert dict to object."""
        return cls(StringOpts.from_config(config.options()))

    @classmethod
    def commands_type(cls) -> type[bash.OnlyOptions]:
        """Get commands type."""
        # DOCU user can change CommandsOnlyOptions
        return bash.OnlyOptions

    def sh_commands(
        self,
        exp_fs_managers: exp_fs.Managers,
    ) -> bash.OnlyOptions:
        """Get sh commands."""
        return self.commands_type()(
            bash.Options(self._options),
            exp_fs_managers,
            self.parent_dir_where_defined(),
        )

    def to_config(self) -> cfg.OnlyOptions:
        """Convert to dict."""
        return cfg.OnlyOptions(self._options.to_config())


class WithArguments[N: Names](WithOptions[cfg.WithArguments, ArgsLoadError]):
    """Tool config with arguments."""

    @classmethod
    @abstractmethod
    def arguments_type(cls) -> type[Arguments[N]]:
        """Get argument arguments type."""
        raise NotImplementedError

    @classmethod
    def config_type(cls) -> type[cfg.WithArguments]:
        """Get config type."""
        return cfg.WithArguments

    @classmethod
    def from_config(cls, config: cfg.WithArguments) -> Self | ArgsLoadError:
        """Convert dict to object."""
        match arg_or_err := cls.arguments_type().from_config(config.arguments()):
            case Arguments():
                return cls(arg_or_err, StringOpts.from_config(config.options()))
            case _:
                return arg_or_err

    def __init__(self, arguments: Arguments[N], options: StringOpts) -> None:
        """Initialize."""
        super().__init__(options)
        self._arguments = arguments

    def arguments(self) -> Arguments[N]:
        """Get arguments."""
        return self._arguments

    def sh_commands(
        self,
        exp_fs_managers: exp_fs.Managers,
    ) -> bash.WithArguments:
        """Get sh commands."""
        return bash.WithArguments(
            self._arguments.sh_lines_builders(exp_fs_managers),
            self.options().sh_lines_builder(),
            exp_fs_managers,
            self.parent_dir_where_defined(),
        )

    def to_config(self) -> cfg.WithArguments:
        """Convert to dict."""
        return cfg.WithArguments(
            self._arguments.to_config(),
            self._options.to_config(),
        )


def get_arg[A: Arg](
    exp_fs_manager: exp_fs.ManagerBase,
    arg_type: type[A],
) -> A | InvalidToolNameError:
    """Get argument."""
    arg_config = cfg.Arg.from_yaml(exp_fs_manager.config_yaml())
    return arg_type.from_config(arg_config)



"""Experiment SLURM step checking module."""

import slurmbench.experiment.bash.items as exp_bash_items
import slurmbench.experiment.file_system as exp_fs
import slurmbench.experiment.slurm.status as exp_slurm_status


def script_step_status(
    work_exp_fs_manager: exp_fs.WorkManager,
    job_id: str,
    step: exp_bash_items.Steps,
) -> exp_slurm_status.ScriptSteps:
    """Get SLURM script step status thanks to the experiment file system."""
    ok_file = work_exp_fs_manager.slurm_log_fs_manager().script_step_status_file(
        job_id,
        step,
        exp_slurm_status.ScriptSteps.OK,
    )
    error_file = work_exp_fs_manager.slurm_log_fs_manager().script_step_status_file(
        job_id,
        step,
        exp_slurm_status.ScriptSteps.ERROR,
    )
    if ok_file.exists():
        status = exp_slurm_status.ScriptSteps.OK
    elif error_file.exists():
        status = exp_slurm_status.ScriptSteps.ERROR
    else:
        status = exp_slurm_status.ScriptSteps.NULL
    ok_file.unlink(missing_ok=True)
    error_file.unlink(missing_ok=True)
    return status



"""Sample slurm file system manager."""

from __future__ import annotations

from pathlib import Path

import slurmbench.samples.slurm.status as smp_slurm_status
from slurmbench.slurm import sacct


class Manager:
    """Sample slurm file system manager."""

    STATS_PSV_NAME = Path("stats.psv")
    STDOUT_LOG_NAME = Path("stdout.log")
    STDERR_LOG_NAME = Path("stderr.log")

    def __init__(self, slurm_directory: Path) -> None:
        """Initialize."""
        self.__slurm_dir = slurm_directory
        self.__command_steps_status_file_manager = CommandStepStatusFileManager(
            self.__slurm_dir,
        )
        self.__job_state_file_builder = JobStateFileManager(self.__slurm_dir)

    def slurm_dir(self) -> Path:
        """Get slurm sample directory path."""
        return self.__slurm_dir

    def command_steps_status_file_manager(self) -> CommandStepStatusFileManager:
        """Get command steps status file manager."""
        return self.__command_steps_status_file_manager

    def stats_psv(self) -> Path:
        """Get stats file path."""
        return self.__slurm_dir / self.STATS_PSV_NAME

    def stdout_log(self) -> Path:
        """Get stdout file path."""
        return self.__slurm_dir / self.STDOUT_LOG_NAME

    def stderr_log(self) -> Path:
        """Get stderr file path."""
        return self.__slurm_dir / self.STDERR_LOG_NAME

    def job_state_file_builder(self) -> JobStateFileManager:
        """Get job state file builder."""
        return self.__job_state_file_builder


class CommandStepStatusFileManager:
    """Command step status file manager."""

    YAML_NAME = "command_steps_status.yaml"

    def __init__(self, directory: Path) -> None:
        """Initialize."""
        self.__directory = directory

    def path(self) -> Path:
        """Get command step status file."""
        return self.__directory / self.YAML_NAME

    def get_command_step_status(self) -> smp_slurm_status.CommandStepsProcess:
        """Get command step status."""
        return smp_slurm_status.CommandStepsProcess.from_yaml(self.path())


class JobStateFileManager:
    """Job state file builder."""

    PREFIX = "job_state"

    @classmethod
    def filename(cls, sacct_state: sacct.State) -> Path:
        """Build job state file."""
        return Path(f"{cls.PREFIX}.{sacct_state}")

    def __init__(self, directory: Path) -> None:
        """Initialize."""
        self.__directory = directory

    def path(self, sacct_state: sacct.State) -> Path:
        """Get job state file."""
        return self.__directory / self.filename(sacct_state)

    def find_path(self) -> Path:
        """Get job state file.

        Raises
        ------
        FileNotFoundError
            If job state file is not found.
        """
        path = next(
            (
                file
                for file in self.__directory.iterdir()
                if file.name.startswith(self.PREFIX)
            ),
            None,
        )
        if path is None:
            _error_msg = f"Job state file not found in {self.__directory}"
            raise FileNotFoundError(_error_msg)

        return path

    def extract_sacct_state(self) -> sacct.State:
        """Get job state."""
        return sacct.State(self.find_path().suffix)



"""Topic visitor abstract module."""

from __future__ import annotations

import logging
from abc import abstractmethod
from enum import StrEnum
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    import slurmbench.tool.description as tool_desc


_LOGGER = logging.getLogger(__name__)


class Tools(StrEnum):
    """Topic tools."""

    @classmethod
    def from_description(
        cls,
        tool_description: tool_desc.Description,
    ) -> Self:
        """Create tool from description.

        Raises
        ------
        ValueError
            The topic does not have a tool named `tool_name`.
        """
        try:
            return cls(tool_description.name())
        except ValueError as exc:
            _LOGGER.exception(
                "Unknown tool `%s` for topic `%s`",
                tool_description.name(),
                tool_description.name(),
            )
            raise ValueError from exc

    @abstractmethod
    def to_description(self) -> tool_desc.Description:
        """Get tool description."""
        raise NotImplementedError



"""Experiment report module."""

from __future__ import annotations

from typing import TYPE_CHECKING, Self, final

import slurmbench.samples.status as smp_status
import slurmbench.slurm.bash as slurm_bash
from slurmbench.slurm import sacct
from slurmbench.yaml_interface import YAMLInterface

from . import file_system as exp_fs
from . import in_progress, stats
from . import iter as exp_iter

if TYPE_CHECKING:
    from yaml import YAMLObject  # type: ignore[import-untyped]


@final
class RunningExperiment(YAMLInterface):
    """Running experiment metadata."""

    KEY_IN_PROGRESS_DATA = "in_progress_data"
    KEY_SACCT_STATE = "sacct_state"

    @classmethod
    def from_yaml_load(cls, pyyaml_obj: dict[str, YAMLObject]) -> Self:
        """Convert dict to object."""
        sacct_state = (
            sacct.State(pyyaml_obj[cls.KEY_SACCT_STATE])
            if pyyaml_obj[cls.KEY_SACCT_STATE] is not None
            else None
        )
        return cls(
            in_progress.InDataDirectory.from_yaml_load(
                pyyaml_obj[cls.KEY_IN_PROGRESS_DATA],
            ),
            sacct_state,
        )

    def __init__(
        self,
        in_progress_data: in_progress.InDataDirectory,
        sacct_state: sacct.State | None,
    ) -> None:
        self._in_progress_data = in_progress_data
        self._sacct_state = sacct_state

    def in_progress_data(self) -> in_progress.InDataDirectory:
        """Get in progress metadata in the data directory."""
        return self._in_progress_data

    def sacct_state(self) -> sacct.State | None:
        """Get sacct state."""
        return self._sacct_state

    def to_yaml_dump(self) -> dict[str, YAMLObject]:
        """Convert to dict."""
        return {
            self.KEY_IN_PROGRESS_DATA: self._in_progress_data.to_yaml_dump(),
            self.KEY_SACCT_STATE: (
                self._sacct_state.value if self._sacct_state is not None else None
            ),
        }


def running_experiment_to_str(running_experiment: RunningExperiment) -> str:
    """Convert running experiment to string."""
    return (
        f"Experiment is running\n"
        f"* Experiment launched the: {running_experiment.in_progress_data().date()}\n"
        f"* Working directory root path:"
        f" {running_experiment.in_progress_data().working_directory()}\n"
        f"* SLURM job ID: {running_experiment.in_progress_data().job_id()}\n"
        f"* SACCT state: {running_experiment.sacct_state()}\n"
    )


def running_experiment(
    data_exp_fs_manager: exp_fs.DataManager,
) -> RunningExperiment | None:
    """Check if experiment is running."""
    if not data_exp_fs_manager.in_progress_yaml().exists():
        return None

    in_progress_data = in_progress.InDataDirectory.from_yaml(
        data_exp_fs_manager.in_progress_yaml(),
    )
    return RunningExperiment(
        in_progress_data,
        slurm_bash.get_state(in_progress_data.job_id()),
    )


def samples_stats(data_exp_manager: exp_fs.DataManager) -> stats.Samples:
    """Get samples stats."""
    status_number = smp_status.StatusMap[int].default(0)
    for _, status in exp_iter.samples_with_status(data_exp_manager):
        status_number[status] += 1
    return stats.Samples.from_status_map(status_number)


@final
class Report(YAMLInterface):
    """report metadata."""

    KEY_RUNNING_EXPERIMENT = "running_experiment"
    KEY_SAMPLES_STATS = "samples_stats"

    @classmethod
    def from_data_exp_fs_manager(cls, data_exp_manager: exp_fs.DataManager) -> Self:
        """Generate report."""
        return cls(
            running_experiment(data_exp_manager),
            samples_stats(data_exp_manager),
        )

    @classmethod
    def from_yaml_load(cls, pyyaml_obj: dict[str, YAMLObject]) -> Self:
        """Convert dict to object."""
        running_experiment = (
            RunningExperiment.from_yaml_load(pyyaml_obj[cls.KEY_RUNNING_EXPERIMENT])
            if pyyaml_obj[cls.KEY_RUNNING_EXPERIMENT] is not None
            else None
        )
        return cls(
            running_experiment,
            stats.Samples.from_yaml_load(pyyaml_obj[cls.KEY_SAMPLES_STATS]),
        )

    def __init__(
        self,
        running_experiment: RunningExperiment | None,
        samples_stats: stats.Samples,
    ) -> None:
        self._running_experiment = running_experiment
        self._samples_stats = samples_stats

    def running_experiment(self) -> RunningExperiment | None:
        """Get running experiment."""
        return self._running_experiment

    def samples_stats(self) -> stats.Samples:
        """Get samples stats."""
        return self._samples_stats

    def to_yaml_dump(self) -> dict[str, YAMLObject]:
        """Convert to dict."""
        return {
            self.KEY_RUNNING_EXPERIMENT: (
                self._running_experiment.to_yaml_dump()
                if self._running_experiment is not None
                else None
            ),
            self.KEY_SAMPLES_STATS: self._samples_stats.to_yaml_dump(),
        }


def report_to_string(report: Report) -> str:
    """Convert report to string.

    Notes
    -----
    There is no leading new line.
    """
    running_exp = report.running_experiment()
    return (
        (
            "Experiment is not running\n"
            if running_exp is None
            else f"{running_experiment_to_str(running_exp)}\n"
        )
        + "Samples stats:\n"
        "* Total number of samples:"
        f" {report.samples_stats().total_number_of_samples()}\n"
        "* Number of successful samples:"
        f" {report.samples_stats().number_of_successful_samples()}\n"
        "* Number of samples with missing inputs:"
        f" {report.samples_stats().number_of_samples_with_missing_inputs()}\n"
        "* Number of failed samples:"
        f" {report.samples_stats().number_of_failed_samples()}\n"
        "* Number of not run samples:"
        f" {report.samples_stats().number_of_not_run_samples()}"
    )



"""Autodoc application."""

# Due to typer usage:

from __future__ import annotations

import shutil
from pathlib import Path

import typer

import slurmbench.doc.environments as doc_env

APP = typer.Typer(name="doc", help="Generate documentation", rich_markup_mode="rich")

APP.command(name="tools-envs", help="Example of how to manage the tools environments")(
    doc_env.main,
)

MAIN_DOC_DIR = Path("docs")
ENVIRONMENTS_DOC_DIR = MAIN_DOC_DIR / "environments"


@APP.command()
def clean() -> None:
    """Clean the auto documentation."""
    shutil.rmtree(ENVIRONMENTS_DOC_DIR, ignore_errors=True)


@APP.command()
def auto() -> None:
    """Generate the auto documentation."""
    ENVIRONMENTS_DOC_DIR.mkdir(parents=True, exist_ok=True)
    doc_env.main(ENVIRONMENTS_DOC_DIR)



"""Run SLURM job logics."""

import logging
import subprocess

import typer

import slurmbench.experiment.bash.manager as bash_manager
import slurmbench.slurm.bash as slurm_bash
from slurmbench import subprocess_lib

_LOGGER = logging.getLogger(__name__)


def run(sh_manager: bash_manager.Manager) -> None:
    """Run sbatch script."""
    cmd_path = subprocess_lib.command_path(slurm_bash.SBATCH_CMD)
    result = subprocess.run(  # noqa: S603
        [str(x) for x in [cmd_path, sh_manager.work_sh_fs_manager().sbatch_script()]],
        capture_output=True,
        text=True,
        check=False,
    )
    _LOGGER.debug("%s stdout: %s", slurm_bash.SBATCH_CMD, result.stdout)
    _LOGGER.debug("%s stderr: %s", slurm_bash.SBATCH_CMD, result.stderr)
    if result.returncode != 0:
        _LOGGER.critical("Running sbatch script failed:\n%s", result.stderr)
        raise typer.Exit(1)



"""Samples slurm status logics."""

from typing import Self, final

import slurmbench.experiment.slurm.status as exp_slurm_status
from slurmbench.yaml_interface import YAMLInterface


@final
class CommandStepsProcess(YAMLInterface):
    """Command steps process status."""

    INIT_ENV_KEY = "init_env"
    COMMAND_KEY = "command"
    CLOSE_ENV_KEY = "close_env"

    @classmethod
    def from_yaml_load(cls, pyyaml_obj: dict[str, str]) -> Self:
        """Convert pyyaml object to self."""
        return cls(
            exp_slurm_status.ScriptSteps(pyyaml_obj[cls.INIT_ENV_KEY]),
            exp_slurm_status.ScriptSteps(pyyaml_obj[cls.COMMAND_KEY]),
            exp_slurm_status.ScriptSteps(pyyaml_obj[cls.CLOSE_ENV_KEY]),
        )

    def __init__(
        self,
        init_env: exp_slurm_status.ScriptSteps,
        command: exp_slurm_status.ScriptSteps,
        close_env: exp_slurm_status.ScriptSteps,
    ) -> None:
        """Initialize."""
        self.__init_env = init_env
        self.__command = command
        self.__close_env = close_env

    def init_env(self) -> exp_slurm_status.ScriptSteps:
        """Get init env command step."""
        return self.__init_env

    def command(self) -> exp_slurm_status.ScriptSteps:
        """Get command command step."""
        return self.__command

    def close_env(self) -> exp_slurm_status.ScriptSteps:
        """Get close env command step."""
        return self.__close_env

    def to_yaml_dump(self) -> dict[str, str]:
        """Convert self to pyyaml object."""
        return {
            self.INIT_ENV_KEY: self.__init_env.value,
            self.COMMAND_KEY: self.__command.value,
            self.CLOSE_ENV_KEY: self.__close_env.value,
        }



"""In progress experiment logics."""

from __future__ import annotations

import time
from abc import ABC, abstractmethod
from datetime import UTC, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Self, final

from slurmbench.yaml_interface import YAMLInterface

if TYPE_CHECKING:
    from . import file_system as exp_fs
    from . import managers as exp_managers
    from .bash import manager as bash_manager


class Base(YAMLInterface, ABC):
    """Base class for in progress metadata."""

    KEY_DATE = "date"
    KEY_JOB_ID = "job_id"

    @classmethod
    @abstractmethod
    def root_directory_key(cls) -> str:
        """Get root directory key."""
        raise NotImplementedError

    @classmethod
    def from_yaml_load(cls, obj_dict: dict[str, Any]) -> Self:
        """Convert dict to object."""
        return cls(
            obj_dict[cls.KEY_DATE],
            Path(obj_dict[cls.root_directory_key()]),
            obj_dict[cls.KEY_JOB_ID],
        )

    def __init__(
        self,
        date: str,
        twin_root_directory: Path,
        job_id: str,
    ) -> None:
        self._date = date
        self._twin_root_directory = twin_root_directory
        self._job_id = job_id

    def date(self) -> str:
        """Get date."""
        return self._date

    def twin_root_directory(self) -> Path:
        """Get twin root directory."""
        return self._twin_root_directory

    def job_id(self) -> str:
        """Get job id."""
        return self._job_id

    def to_yaml_dump(self) -> dict[str, Any]:
        """Convert to dict."""
        return {
            self.KEY_DATE: self._date,
            self.root_directory_key(): str(self._twin_root_directory),
            self.KEY_JOB_ID: self._job_id,
        }


@final
class InDataDirectory(Base):
    """In data experiment metadata."""

    @classmethod
    def root_directory_key(cls) -> str:
        """Get root directory key."""
        return "working_directory"

    def working_directory(self) -> Path:
        """Get working directory."""
        return self._twin_root_directory


@final
class InWorkingDirectory(Base):
    """In working experiment metadata."""

    @classmethod
    def root_directory_key(cls) -> str:
        """Get root directory key."""
        return "data_directory"

    def data_directory(self) -> Path:
        """Get data directory."""
        return self._twin_root_directory


def get_today_format_string() -> str:
    """Get date format string."""
    return datetime.now(tz=UTC).strftime("%Y-%m-%d_%H-%M-%S")


def write_in_progress_metadata(
    exp_manager: exp_managers.OnlyOptions | exp_managers.WithArguments,
    sh_manager: bash_manager.Manager,
) -> str:
    """Write in progress metadata."""
    job_id = _get_array_job_id(exp_manager.fs_managers().work())

    data_in_progress = InDataDirectory(
        sh_manager.date_str(),
        exp_manager.fs_managers().work().root_dir(),
        job_id,
    )
    work_in_progress = InWorkingDirectory(
        sh_manager.date_str(),
        exp_manager.fs_managers().data().root_dir(),
        job_id,
    )

    data_in_progress.to_yaml(exp_manager.fs_managers().data().in_progress_yaml())
    work_in_progress.to_yaml(exp_manager.fs_managers().work().in_progress_yaml())

    return job_id


def _get_array_job_id(work_exp_fs_manager: exp_fs.WorkManager) -> str:
    """Wait the tmp array job id file is created and extract the array job id."""
    tmp_job_id_file = (
        work_exp_fs_manager.slurm_log_fs_manager().job_id_file_manager().path()
    )
    while not tmp_job_id_file.exists():
        time.sleep(5)

    job_id = (
        work_exp_fs_manager.slurm_log_fs_manager().job_id_file_manager().get_job_id()
    )
    tmp_job_id_file.unlink()
    return job_id



"""Slurm configs."""

from __future__ import annotations


def default_slurm_options(account_name: str | None) -> str:
    """Return default SLURM options."""
    account_opt = "" if account_name is None else f"--account={account_name}"
    return f"--mem=4096 --cpus-per-task=4 --time=1:00:00 {account_opt}"



"""Sample missing inputs."""

from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING, final

import slurmbench.experiment.managers as exp_managers
import slurmbench.tool.connector as tool_connector
import slurmbench.tool.results as tool_res
from slurmbench import tab_files

from . import file_system as smp_fs
from . import items as smp
from . import status as smp_status

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator


class MissingInput:
    """Missing input item."""

    @classmethod
    def from_tool_input(
        cls,
        arg_name: str,
        tool_input: tool_res.Result,
        reason: smp_status.Error,
        help_string: str,
    ) -> MissingInput:
        """Create missing input from tool input."""
        return cls(
            str(arg_name),
            tool_input.exp_fs_manager().topic_dir().name,
            tool_input.exp_fs_manager().tool_dir().name,
            tool_input.exp_fs_manager().exp_dir().name,
            reason,
            help_string,
        )

    def __init__(  # noqa: PLR0913
        self,
        arg_name: str,
        topic_name: str,
        tool_name: str,
        experiment_name: str,
        reason: smp_status.Error,
        help_string: str,
    ) -> None:
        """Initialize."""
        self.__arg_name = arg_name
        self.__topic_name = topic_name
        self.__tool_name = tool_name
        self.__exp_name = experiment_name
        self.__reason = reason
        self.__help_string = help_string

    def arg_name(self) -> str:
        """Get argument name."""
        return self.__arg_name

    def topic_name(self) -> str:
        """Get topic name."""
        return self.__topic_name

    def tool_name(self) -> str:
        """Get tool name."""
        return self.__tool_name

    def experiment_name(self) -> str:
        """Get experiment name."""
        return self.__exp_name

    def reason(self) -> smp_status.Error:
        """Get reason."""
        return self.__reason

    def help(self) -> str:
        """Get help string."""
        return self.__help_string


class MissingInputsTSVHeader(StrEnum):
    """Missing inputs TSV header."""

    ARG_NAME = "arg_name"
    TOPIC = "input_topic"
    TOOL = "input_tool"
    EXPERIMENT = "input_experiment"
    REASON = "reason"
    HELP = "help"


@final
class MissingInputsTSVReader(tab_files.TSVReader[MissingInputsTSVHeader, MissingInput]):
    """Missing inputs TSV reader."""

    @classmethod
    def header_type(cls) -> type[MissingInputsTSVHeader]:
        """Get header type."""
        return MissingInputsTSVHeader

    def __iter__(self) -> Iterator[MissingInput]:
        """Iterate over missing inputs items."""
        for row in self._csv_reader:
            yield MissingInput(
                self._get_cell(row, MissingInputsTSVHeader.ARG_NAME),
                self._get_cell(row, MissingInputsTSVHeader.TOPIC),
                self._get_cell(row, MissingInputsTSVHeader.TOOL),
                self._get_cell(row, MissingInputsTSVHeader.TOOL),
                smp_status.Error(
                    self._get_cell(row, MissingInputsTSVHeader.REASON),
                ),
                self._get_cell(row, MissingInputsTSVHeader.HELP),
            )


@final
class MissingInputsTSVWriter(tab_files.TSVWriter[MissingInputsTSVHeader, MissingInput]):
    """Missing inputs TSV writer."""

    @classmethod
    def header_type(cls) -> type[MissingInputsTSVHeader]:
        """Get header type."""
        return MissingInputsTSVHeader

    @classmethod
    def reader_type(cls) -> type[MissingInputsTSVReader]:
        """Get reader type."""
        return MissingInputsTSVReader

    def _to_cell(self, item: MissingInput, column_id: MissingInputsTSVHeader) -> object:
        """Get cell from item."""
        match column_id:
            case MissingInputsTSVHeader.ARG_NAME:
                return item.arg_name()
            case MissingInputsTSVHeader.TOPIC:
                return item.topic_name()
            case MissingInputsTSVHeader.TOOL:
                return item.tool_name()
            case MissingInputsTSVHeader.EXPERIMENT:
                return item.experiment_name()
            case MissingInputsTSVHeader.REASON:
                return item.reason()
            case MissingInputsTSVHeader.HELP:
                return item.help()


def write_sample_missing_inputs(
    exp_manager: exp_managers.WithArguments,
    row_numbered_sample: smp.RowNumbered,
    sample_missing_inputs: Iterable[MissingInput],
) -> None:
    """Write sample missing inputs."""
    data_sample_fs_manager = (
        exp_manager.fs_managers()
        .data()
        .sample_fs_manager(
            row_numbered_sample,
        )
    )
    smp_fs.reset_sample_dir(data_sample_fs_manager)
    with MissingInputsTSVWriter.open(
        data_sample_fs_manager.missing_inputs_tsv(),
        "w",
    ) as out_miss_inputs:
        out_miss_inputs.write_bunch(sample_missing_inputs)


def for_sample[N: tool_connector.Names](
    tool_inputs: dict[N, tool_res.Result],
    sample: smp.RowNumbered,
) -> list[MissingInput]:
    """Get a list of missing inputs."""
    list_missing_inputs: list[MissingInput] = []
    for arg_name, tool_input in tool_inputs.items():
        input_status = tool_input.check(sample)
        match input_status:
            case smp_status.Error():
                list_missing_inputs.append(
                    MissingInput.from_tool_input(
                        str(arg_name),
                        tool_input,
                        input_status,
                        _get_help_str(tool_input),
                    ),
                )
    return list_missing_inputs


def _get_help_str(tool_input: tool_res.Result) -> str:
    """Get help string."""
    return (
        "slurmbench"
        f" {tool_input.exp_fs_manager().tool_description().topic().cmd()}"
        f" {tool_input.exp_fs_manager().tool_description().cmd()}"
        f" run"
        " --help"
    )



"""Interface for topics visitor."""

from __future__ import annotations

import logging
from abc import abstractmethod
from enum import StrEnum
from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    import slurmbench.topic.description as topic_desc
    import slurmbench.topic.visitor as topic_visitor

_LOGGER = logging.getLogger(__name__)


class Topics(StrEnum):
    """Topics."""

    @classmethod
    def from_description(
        cls,
        topic_description: topic_desc.Description,
    ) -> Self:
        """Create topic from description.

        Raises
        ------
        ValueError
            The topic does not have a tool named `tool_name`.
        """
        try:
            return cls(topic_description.name())
        except ValueError as exc:
            _LOGGER.exception(
                "Unknown topic `%s`",
                topic_description.name(),
            )
            raise ValueError from exc

    @abstractmethod
    def to_description(self) -> topic_desc.Description:
        """Get topic description."""
        raise NotImplementedError

    @abstractmethod
    def tools(self) -> type[topic_visitor.Tools]:
        """Get tools."""
        raise NotImplementedError
